@@ -38,7 +38,7 @@ class Imagine:
         self.half_mode = half_mode
         self.precision = precision
         model_config = config.get_model_config(model_name)
-        control_model_config = config.get_control_model_config(self.control_model_name)
+        control_model_config = config.get_control_model_config(self.control_model_name) if self.control_model_name else None
 
         if model_config is None:
             raise ValueError(f"Unknown model name: {model_name}")
@@ -46,7 +46,7 @@ class Imagine:
         self.sd_model = get_diffusion_model(
             weights_location=model_config.weights_url,
             config_path=model_config.config_path,
-            control_weights_location=control_model_config.weights_url,
+            control_weights_location=control_model_config.weights_url if control_model_config else None,
             half_mode=half_mode,
             for_inpainting=for_inpainting,
         )
