@@ -1,9 +1,11 @@
 import logging
+import math
 import os
 import re
 
 from classy_imaginary.schema import SafetyMode
 from classy_imaginary.model_manager import get_diffusion_model
+from classy_imaginary.img_utils import add_caption_to_image
 
 logger = logging.getLogger(__name__)
 
@@ -215,7 +217,7 @@ def imagine(
                     half_mode=half_mode,
                     add_caption=add_caption,
                 )
-                if not result.is_nsfw:
+                if not result.safety_score.is_filtered:
                     break
                 if attempt < unsafe_retry_count:
                     logger.info("   Image was unsafe, retrying with new seed...")
@@ -226,7 +228,6 @@ def imagine(
 def _generate_single_image(
         prompt,
         model,
-        midas_model=None,
         nsfw_filter: bool = False,
         safety_feature_extractor=None,
         safety_checker=None,
@@ -235,27 +236,29 @@ def _generate_single_image(
         progress_img_callback=None,
         progress_img_interval_steps=3,
         progress_img_interval_min_s=0.1,
-        half_mode=None,
         add_caption=False,
+        suppress_inpaint=False,
+        return_latent=False,
 ):
-    import numpy as np
     import torch.nn
-    from einops import rearrange, repeat
     from PIL import Image, ImageOps
     from pytorch_lightning import seed_everything
-    from torch.cuda import OutOfMemoryError
 
     from classy_imaginary.enhancers.clip_masking import get_img_mask
     from classy_imaginary.enhancers.describe_image_blip import generate_caption
     from classy_imaginary.enhancers.face_restoration_codeformer import enhance_faces
     from classy_imaginary.enhancers.upscale_realesrgan import upscale_image
-    from classy_imaginary.img_utils import pillow_fit_image_within, pillow_img_to_torch_image
+    from classy_imaginary.img_utils import (pillow_fit_image_within,
+        pillow_img_to_torch_image,
+        pillow_mask_to_latent_mask,
+        torch_img_to_pillow_img)
     from classy_imaginary.log_utils import (
         ImageLoggingContext,
         log_conditioning,
         log_img,
         log_latent,
     )
+    from classy_imaginary.modules.midas.api import torch_image_to_depth_map
     from classy_imaginary.outpaint import outpaint_arg_str_parse, prepare_image_for_outpaint
     from classy_imaginary.safety import create_safety_score
     from classy_imaginary.samplers import SAMPLER_LOOKUP
@@ -276,7 +279,8 @@ def _generate_single_image(
         _, img_type = prompt.mask_image.strip("*").split(".")
         prompt.mask_image = _most_recent_result.images[img_type]
 
-    has_depth_channel = hasattr(model, "depth_stage_key")
+    is_controlnet_model = hasattr(model, "control_key")
+
     progress_latents = []
 
     def latent_logger(latents):
@@ -299,6 +303,7 @@ def _generate_single_image(
         with lc.timing("conditioning"):
             # need to expand if doing batches
             neutral_conditioning = _prompts_to_embeddings(prompt.negative_prompt, model)
+            _prompts_to_embeddings("", model)
             log_conditioning(neutral_conditioning, "neutral conditioning")
             if prompt.conditioning is not None:
                 positive_conditioning = prompt.conditioning
@@ -314,9 +319,17 @@ def _generate_single_image(
         ]
         SamplerCls = SAMPLER_LOOKUP[prompt.sampler_type.lower()]
         sampler = SamplerCls(model)
-        mask = mask_image = mask_image_orig = mask_grayscale = None
-        t_enc = init_latent = init_latent_noised = None
+        mask_latent = mask_image = mask_image_orig = mask_grayscale = None
+        t_enc = init_latent = init_latent_noised = control_image = None
         starting_image = None
+        denoiser_cls = None
+
+        c_cat = []
+        c_cat_neutral = None
+        result_images = {}
+        seed_everything(prompt.seed)
+        noise = randn_seeded(seed=prompt.seed, size=shape).to(get_device())
+
         if prompt.init_image:
             starting_image = prompt.init_image
             generation_strength = 1 - prompt.init_image_strength
@@ -356,25 +369,12 @@ def _generate_single_image(
                 if prompt.mask_mode == ImaginePrompt.MaskMode.REPLACE:
                     mask_image = ImageOps.invert(mask_image)
 
-                log_img(
-                    Image.composite(init_image, mask_image, mask_image),
-                    "mask overlay",
-                )
                 mask_image_orig = mask_image
-                mask_image = mask_image.resize(
-                    (
-                        mask_image.width // downsampling_factor,
-                        mask_image.height // downsampling_factor,
-                    ),
-                    resample=Image.Resampling.LANCZOS,
-                )
                 log_img(mask_image, "latent_mask")
+                mask_latent = pillow_mask_to_latent_mask(
+                    mask_image, downsampling_factor=downsampling_factor
+                ).to(get_device())
 
-                mask = np.array(mask_image)
-                mask = mask.astype(np.float32) / 255.0
-                mask = mask[None, None]
-                mask = torch.from_numpy(mask)
-                mask = mask.to(get_device())
             init_image_t = pillow_img_to_torch_image(init_image)
             init_image_t = init_image_t.to(get_device())
             init_latent = model.get_first_stage_encoding(
@@ -383,10 +383,11 @@ def _generate_single_image(
             shape = init_latent.shape
 
             log_latent(init_latent, "init_latent")
-            # encode (scaled latent)
             seed_everything(prompt.seed)
-            noise = randn_seeded(seed=prompt.seed, size=init_latent.size())
-            noise = noise.to(get_device())
+            noise = randn_seeded(seed=prompt.seed, size=init_latent.shape).to(
+                get_device()
+            )
+            # noise = noise[:, :, : init_latent.shape[2], : init_latent.shape[3]]
 
             schedule = NoiseSchedule(
                 model_num_timesteps=model.num_timesteps,
@@ -406,89 +407,81 @@ def _generate_single_image(
                     schedule=schedule,
                     noise=noise,
                 )
-        batch_size = 1
-        log_latent(init_latent_noised, "init_latent_noised")
-        batch = {
-            "txt": batch_size * [prompt.prompt_text],
-        }
-        c_cat = []
-        c_cat_neutral = None
-        depth_image_display = None
-        if has_depth_channel and starting_image:
-            _init_image_d = np.array(starting_image.convert("RGB"))
-            _init_image_d = (
-                    torch.from_numpy(_init_image_d).to(dtype=torch.float32) / 127.5 - 1.0
-            )
-            depth_image = midas_model(_init_image_d)
-            depth_image = torch.from_numpy(depth_image[None, ...])
-            batch[model.depth_stage_key] = depth_image.to(device=get_device())
-            _init_image_d = rearrange(_init_image_d, "h w c -> 1 c h w")
-            batch["jpg"] = _init_image_d
-            for ck in model.concat_keys:
-                cc = batch[ck]
-                cc = model.depth_model(cc)
-                depth_min, depth_max = torch.amin(
-                    cc, dim=[1, 2, 3], keepdim=True
-                ), torch.amax(cc, dim=[1, 2, 3], keepdim=True)
-                display_depth = (cc - depth_min) / (depth_max - depth_min)
-                depth_image_display = Image.fromarray(
-                    (display_depth[0, 0, ...].cpu().numpy() * 255.0).astype(np.uint8)
-                )
-                cc = torch.nn.functional.interpolate(
-                    cc,
+
+            if hasattr(model, "depth_stage_key"):
+                # depth model
+                depth_t = torch_image_to_depth_map(init_image_t)
+                depth_latent = torch.nn.functional.interpolate(
+                    depth_t,
                     size=shape[2:],
                     mode="bicubic",
                     align_corners=False,
                 )
-                depth_min, depth_max = torch.amin(
-                    cc, dim=[1, 2, 3], keepdim=True
-                ), torch.amax(cc, dim=[1, 2, 3], keepdim=True)
-                cc = 2.0 * (cc - depth_min) / (depth_max - depth_min) - 1.0
-                c_cat.append(cc)
-            c_cat = [torch.cat(c_cat, dim=1)]
+                result_images["depth_image"] = depth_t
+                c_cat.append(depth_latent)
 
-        if mask_image_orig and not has_depth_channel:
-            mask_t = pillow_img_to_torch_image(ImageOps.invert(mask_image_orig)).to(
-                get_device()
-            )
-            inverted_mask = 1 - mask
-            masked_image_t = init_image_t * (mask_t < 0.5)
-            batch.update(
-                {
-                    "image": repeat(
-                        init_image_t.to(device=get_device()),
-                        "1 ... -> n ...",
-                        n=batch_size,
-                    ),
-                    "txt": batch_size * [prompt.prompt_text],
-                    "mask": repeat(
-                        inverted_mask.to(device=get_device()),
-                        "1 ... -> n ...",
-                        n=batch_size,
-                    ),
-                    "masked_image": repeat(
-                        masked_image_t.to(device=get_device()),
-                        "1 ... -> n ...",
-                        n=batch_size,
-                    ),
-                }
-            )
+            elif is_controlnet_model and starting_image:
+                from classy_imaginary.img_processors.control_modes import CONTROL_MODES
 
-            for concat_key in getattr(model, "concat_keys", []):
-                cc = batch[concat_key].float()
-                if concat_key != model.masked_image_key:
-                    bchw = [batch_size, 4, shape[2], shape[3]]
-                    cc = torch.nn.functional.interpolate(cc, size=bchw[-2:])
-                else:
-                    cc = model.get_first_stage_encoding(model.encode_first_stage(cc))
-                c_cat.append(cc)
+                control_image_input = pillow_fit_image_within(
+                    prompt.control_image,
+                    max_height=prompt.height,
+                    max_width=prompt.width,
+                )
+                control_image_input_t = pillow_img_to_torch_image(control_image_input)
+                control_image_input_t = control_image_input_t.to(get_device())
+
+                control_image = CONTROL_MODES[prompt.control_mode](
+                    control_image_input_t
+                )
+                if len(control_image.shape) == 3:
+                    raise RuntimeError("Control image must be 4D")
+
+                if control_image.shape[1] != 3:
+                    raise RuntimeError("Control image must have 3 channels")
+
+                if control_image.min() < 0 or control_image.max() > 1:
+                    raise RuntimeError(
+                        f"Control image must be in [0, 1] but we received {control_image.min()} and {control_image.max()}"
+                    )
+
+                if control_image.max() <= 0.5:
+                    raise RuntimeError("Control image must be in [0, 1]")
+
+                if control_image.min() >= 0.5:
+                    raise RuntimeError("Control image must be in [0, 1]")
+
+                control_image_disp = control_image * 2 - 1
+                result_images["control"] = control_image_disp[:, [2, 0, 1], :, :]
+                log_img(control_image_disp, "control_image")
+                c_cat.append(control_image)
+
+            elif hasattr(model, "masked_image_key"):
+                # inpainting model
+                mask_t = pillow_img_to_torch_image(ImageOps.invert(mask_image_orig)).to(
+                    get_device()
+                )
+                inverted_mask = 1 - mask_latent
+                masked_image_t = init_image_t * (mask_t < 0.5)
+                log_img(masked_image_t, "masked_image")
+
+                inverted_mask_latent = torch.nn.functional.interpolate(
+                    inverted_mask, size=shape[-2:]
+                )
+                c_cat.append(inverted_mask_latent)
+
+                masked_image_latent = model.get_first_stage_encoding(
+                    model.encode_first_stage(masked_image_t)
+                )
+                c_cat.append(masked_image_latent)
+
+            elif model.cond_stage_key == "edit":
+                # pix2pix model
+                c_cat = [model.encode_first_stage(init_image_t)]
+                c_cat_neutral = [torch.zeros_like(init_latent)]
+                denoiser_cls = CFGEditingDenoiser
             if c_cat:
                 c_cat = [torch.cat(c_cat, dim=1)]
-        denoiser_cls = None
-        if model.cond_stage_key == "edit":
-            c_cat = [model.encode_first_stage(init_image_t).mode()]
-            c_cat_neutral = [torch.zeros_like(init_latent)]
-            denoiser_cls = CFGEditingDenoiser
 
         if c_cat_neutral is None:
             c_cat_neutral = c_cat
@@ -501,6 +494,48 @@ def _generate_single_image(
             "c_concat": c_cat_neutral,
             "c_crossattn": [neutral_conditioning],
         }
+        log_latent(init_latent_noised, "init_latent_noised")
+
+        if (
+            prompt.allow_compose_phase
+            and not is_controlnet_model
+            and not model.cond_stage_key == "edit"
+        ):
+            comp_samples = _generate_composition_latent(
+                sampler=sampler,
+                sampler_kwargs={
+                    "num_steps": prompt.steps,
+                    "initial_latent": init_latent_noised,
+                    "positive_conditioning": positive_conditioning,
+                    "neutral_conditioning": neutral_conditioning,
+                    "guidance_scale": prompt.prompt_strength,
+                    "t_start": t_enc,
+                    "mask": mask_latent,
+                    "orig_latent": init_latent,
+                    "shape": shape,
+                    "batch_size": 1,
+                    "denoiser_cls": denoiser_cls,
+                },
+            )
+            if comp_samples is not None:
+                result_images["composition"] = comp_samples
+                noise = noise[:, :, : comp_samples.shape[2], : comp_samples.shape[3]]
+
+                schedule = NoiseSchedule(
+                    model_num_timesteps=model.num_timesteps,
+                    ddim_num_steps=prompt.steps,
+                    model_alphas_cumprod=model.alphas_cumprod,
+                    ddim_discretize="uniform",
+                )
+                t_enc = int(prompt.steps * 0.75)
+                init_latent_noised = noise_an_image(
+                    comp_samples,
+                    torch.tensor([t_enc - 1]).to(get_device()),
+                    schedule=schedule,
+                    noise=noise,
+                )
+
+                log_latent(comp_samples, "comp_samples")
         with lc.timing("sampling"):
             samples = sampler.sample(
                 num_steps=prompt.steps,
@@ -509,54 +544,44 @@ def _generate_single_image(
                 neutral_conditioning=neutral_conditioning,
                 guidance_scale=prompt.prompt_strength,
                 t_start=t_enc,
-                mask=mask,
+                mask=mask_latent,
                 orig_latent=init_latent,
                 shape=shape,
                 batch_size=1,
                 denoiser_cls=denoiser_cls,
             )
-        # from torch.nn.functional import interpolate
-        # samples = interpolate(samples, scale_factor=2, mode='nearest')
+        if return_latent:
+            return samples
+
         with lc.timing("decoding"):
-            try:
-                x_samples = model.decode_first_stage(samples)
-            except OutOfMemoryError:
-                model.cond_stage_model.to("cpu")
-                model.model.to("cpu")
-                x_samples = model.decode_first_stage(samples)
-                model.cond_stage_model.to(get_device())
-                model.model.to(get_device())
+            gen_imgs_t = model.decode_first_stage(samples)
+            gen_img = torch_img_to_pillow_img(gen_imgs_t)
 
-            x_samples = torch.clamp((x_samples + 1.0) / 2.0, min=0.0, max=1.0)
+        if mask_image_orig and init_image:
+            mask_final = mask_image_orig.copy()
+            log_img(mask_final, "reconstituting mask")
+            mask_final = ImageOps.invert(mask_final)
+            gen_img = Image.composite(gen_img, init_image, mask_final)
+            gen_img = combine_image(
+                original_img=init_image,
+                generated_img=gen_img,
+                mask_img=mask_image_orig,
+            )
+            log_img(gen_img, "reconstituted image")
 
-        for x_sample in x_samples:
-            x_sample = x_sample.to(torch.float32)
-            x_sample = 255.0 * rearrange(x_sample.cpu().numpy(), "c h w -> h w c")
-            x_sample_8_orig = x_sample.astype(np.uint8)
-            img = Image.fromarray(x_sample_8_orig)
-            if mask_image_orig and init_image:
-                # mask_final = mask_image_orig.filter(
-                #     ImageFilter.GaussianBlur(radius=3)
-                # )
-                mask_final = mask_image_orig.copy()
-                log_img(mask_final, "reconstituting mask")
-                mask_final = ImageOps.invert(mask_final)
-                img = Image.composite(img, init_image, mask_final)
-                log_img(img, "reconstituted image")
+        upscaled_img = None
+        rebuilt_orig_img = None
 
-            upscaled_img = None
-            rebuilt_orig_img = None
-
-            if add_caption:
-                caption = generate_caption(img)
-                logger.info(f"Generated caption: {caption}")
+        if add_caption:
+            caption = generate_caption(gen_img)
+            logger.info(f"Generated caption: {caption}")
 
             if nsfw_filter:
                 with lc.timing("safety-filter"):
                     safety_score = create_safety_score(
                         safety_checker,
                         safety_feature_extractor,
-                        img,
+                        gen_img,
                         safety_mode=IMAGINAIRY_SAFETY_MODE,
                     )
             if nsfw_filter and safety_score.is_filtered:
@@ -565,40 +590,28 @@ def _generate_single_image(
                 if prompt.fix_faces and codformer_model:
                     logger.info("Fixing 😊 's in 🖼  using CodeFormer...")
                     with lc.timing("face enhancement"):
-                        img = enhance_faces(img,
+                        gen_img = enhance_faces(gen_img,
                                             net=codformer_model,
                                             fidelity=prompt.fix_faces_fidelity)
                 if prompt.upscale:
                     logger.info("Upscaling 🖼  using real-ESRGAN...")
                     with lc.timing("upscaling"):
-                        upscaled_img = upscale_image(img)
+                        upscaled_img = upscale_image(gen_img)
 
-                # put the newly generated patch back into the original, full size image
-                if prompt.mask_modify_original and mask_image_orig and starting_image:
-                    img_to_add_back_to_original = upscaled_img if upscaled_img else img
-                    img_to_add_back_to_original = img_to_add_back_to_original.resize(
-                        starting_image.size,
-                        resample=Image.Resampling.LANCZOS,
-                    )
+            # put the newly generated patch back into the original, full-size image
+            if prompt.mask_modify_original and mask_image_orig and starting_image:
+                img_to_add_back_to_original = upscaled_img if upscaled_img else gen_img
+                rebuilt_orig_img = combine_image(
+                    original_img=starting_image,
+                    generated_img=img_to_add_back_to_original,
+                    mask_img=mask_image_orig,
+                )
 
-                    mask_for_orig_size = mask_image_orig.resize(
-                        starting_image.size,
-                        resample=Image.Resampling.LANCZOS,
-                    )
-                    # mask_for_orig_size = mask_for_orig_size.filter(
-                    #     ImageFilter.GaussianBlur(radius=5)
-                    # )
-                    log_img(mask_for_orig_size, "mask for original image size")
-
-                    rebuilt_orig_img = Image.composite(
-                        starting_image,
-                        img_to_add_back_to_original,
-                        mask_for_orig_size,
-                    )
-                    log_img(rebuilt_orig_img, "reconstituted original")
+            if prompt.caption_text:
+                add_caption_to_image(gen_img, prompt.caption_text)
 
             result = ImagineResult(
-                img=img,
+                img=gen_img,
                 prompt=prompt,
                 upscaled_img=upscaled_img,
                 is_nsfw=False if nsfw_filter is False else safety_score.is_nsfw,
@@ -606,7 +619,7 @@ def _generate_single_image(
                 modified_original=rebuilt_orig_img,
                 mask_binary=mask_image_orig,
                 mask_grayscale=mask_grayscale,
-                depth_image=depth_image_display,
+                result_images=result_images,
                 timings=lc.get_timings(),
                 progress_latents=progress_latents.copy(),
             )
@@ -624,5 +637,109 @@ def _prompts_to_embeddings(prompts, model):
     return conditioning
 
 
+def calc_scale_to_fit_within(
+    height,
+    width,
+    max_size,
+):
+    if max(height, width) < max_size:
+        return 1
+
+    if width > height:
+        return max_size / width
+
+    return max_size / height
+
+
+def _generate_composition_latent(
+    sampler,
+    sampler_kwargs,
+):
+    from copy import deepcopy
+
+    from torch.nn import functional as F
+
+    from classy_imaginary.enhancers.upscale_riverwing import upscale_latent
+
+    b, c, h, w = orig_shape = sampler_kwargs["shape"]
+    max_compose_gen_size = 768
+    shrink_scale = calc_scale_to_fit_within(
+        height=h,
+        width=w,
+        max_size=int(math.ceil(max_compose_gen_size / 8)),
+    )
+    if shrink_scale >= 1:
+        return None
+
+    new_kwargs = deepcopy(sampler_kwargs)
+
+    # shrink everything
+    new_shape = b, c, int(round(h * shrink_scale)), int(round(w * shrink_scale))
+    initial_latent = new_kwargs["initial_latent"]
+    if initial_latent is not None:
+        initial_latent = F.interpolate(initial_latent, size=new_shape[2:], mode="area")
+
+    for cond in [
+        new_kwargs["positive_conditioning"],
+        new_kwargs["neutral_conditioning"],
+    ]:
+        cond["c_concat"] = [
+            F.interpolate(c, size=new_shape[2:], mode="area") for c in cond["c_concat"]
+        ]
+
+    mask_latent = new_kwargs["mask"]
+    if mask_latent is not None:
+        mask_latent = F.interpolate(mask_latent, size=new_shape[2:], mode="area")
+
+    orig_latent = new_kwargs["orig_latent"]
+    if orig_latent is not None:
+        orig_latent = F.interpolate(orig_latent, size=new_shape[2:], mode="area")
+    t_start = new_kwargs["t_start"]
+    if t_start is not None:
+        gen_strength = new_kwargs["t_start"] / new_kwargs["num_steps"]
+        t_start = int(round(15 * gen_strength))
+    new_kwargs.update(
+        {
+            "num_steps": 15,
+            "initial_latent": initial_latent,
+            "t_start": t_start,
+            "mask": mask_latent,
+            "orig_latent": orig_latent,
+            "shape": new_shape,
+        }
+    )
+    samples = sampler.sample(**new_kwargs)
+
+    samples = upscale_latent(samples)
+    samples = F.interpolate(samples, size=orig_shape[2:], mode="bilinear")
+    return samples
+
+
 def prompt_normalized(prompt):
     return re.sub(r"[^a-zA-Z0-9.,\[\]-]+", "_", prompt)[:130]
+
+
+def combine_image(original_img, generated_img, mask_img):
+    """Combine the generated image with the original image using the mask image."""
+    from PIL import Image
+
+    from classy_imaginary.log_utils import log_img
+
+    generated_img = generated_img.resize(
+        original_img.size,
+        resample=Image.Resampling.LANCZOS,
+    )
+
+    mask_for_orig_size = mask_img.resize(
+        original_img.size,
+        resample=Image.Resampling.LANCZOS,
+    )
+    log_img(mask_for_orig_size, "mask for original image size")
+
+    rebuilt_orig_img = Image.composite(
+        original_img,
+        generated_img,
+        mask_for_orig_size,
+    )
+    log_img(rebuilt_orig_img, "reconstituted original")
+    return rebuilt_orig_img
