@@ -199,7 +199,7 @@ class CLIPDenseBase(nn.Module):
 
         with torch.no_grad():
 
-            # x_inp.shape[2:]
+            x_inp.shape[2:]
 
             if self.n_tokens is not None:
                 stride2 = x_inp.shape[2] // self.n_tokens
